# -*- coding: utf-8 -*-

#  PycaWM
#  menu.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('menu module needs at least Python 2.5!')

import time

from Xlib import X, XK, Xutil, display, protocol

from client import Client
from hookmanager import (add_pre_hook, remove_pre_hook,
                         add_instance_overriding_hook,
                         remove_instance_overriding_hook)
from misc import modifiers, modifier_indexes, good_key

class Menu(object):
    # entry type
    separator, python, shell, sub_menu = range(4)
    up = XK.string_to_keysym('Up')
    down = XK.string_to_keysym('Down')
    enter = XK.string_to_keysym('Return')
    tab = XK.string_to_keysym('Tab')
    escape = XK.string_to_keysym('Escape')
    def __init__(self, wm, x, y, elements):
        self.elements = list(elements)
        if not elements:
            return
        self.current = -1
        self.wm = wm
        self.x = x
        self.y = y
        # the menu won't get a MapRequest, so it won't be managed
        # but that's exactly what we need
        self.display = wm.display
        self.screen = self.display.screen()
        self.root = self.screen.root
        self.font = self.display.open_font('fixed')
        self.heights = list()
        self.width = 0
        for i, element in enumerate(self.elements):
            if element[1] == self.separator and not element[0]:
                self.heights.append(4)
            else:
                element = list(element)
                element[0] = str(element[0]) if element[0] else '(No title)'
                self.elements[i] = element
                extents = self.font.query_text_extents(element[0])
                self.heights.append(extents.overall_ascent + 2)
                self.width = max(self.width, extents.overall_width)
        self.heigth = sum(self.heights)
        self.width += 1
        self.window = self.root.create_window(
            self.x, self.y,
            self.width, self.heigth,
            1, self.screen.root_depth,
            window_class=X.InputOutput,
            event_mask= X.ButtonReleaseMask | X.ExposureMask |
            X.KeyPressMask | X.KeyReleaseMask | X.PointerMotionMask,
            background_pixel=self.screen.black_pixel,
            border_pixel=self.wm.get_color('#5d1a00'),
            save_under=True)
        self.protocols = self.window.get_wm_protocols()
        geo = self.window.get_geometry()
        self.window_pixmap = self.window.create_pixmap(geo.width, geo.height,
                                                       self.screen.root_depth)
        self.gray_gc = self.window.create_gc(
            foreground=self.wm.get_color('#707070'),
            background=self.wm.get_color('#707070'), font=self.font)
        self.darkgray_gc = self.window.create_gc(
            foreground=self.wm.get_color('#606060'),
            background=self.wm.get_color('#606060'), font=self.font)
        self.black_gc = self.window.create_gc(
            foreground=self.screen.black_pixel,
            background=self.wm.get_color('#606060'), font=self.font)
        self.white_gc = self.window.create_gc(
            foreground=self.screen.white_pixel,
            background=self.wm.get_color('#707070'), font=self.font)
        add_pre_hook(wm.expose_reply, self.expose_reply)
        add_pre_hook(wm.motion_reply, self.motion_reply)
        add_pre_hook(wm.button_release_reply, self.button_release_reply)
        add_pre_hook(wm.key_press_reply, self.key_press_reply)
        self.window.map()
        self.window.set_wm_state(state=Xutil.NormalState, icon=0)
        self.window.set_input_focus(X.RevertToPointerRoot, X.CurrentTime)
        if (self.window.grab_pointer(False,
                                     X.ButtonReleaseMask | X.PointerMotionMask,
                                     X.GrabModeAsync,
                                     X.GrabModeAsync,
                                     X.NONE,
                                     X.NONE,
                                     X.CurrentTime) != X.GrabSuccess):
            print 'cannot grab the pointer for the Menu!'
            self.close()
            return
        if (self.window.grab_keyboard(False,
                                      X.GrabModeAsync,
                                      X.GrabModeAsync,
                                      X.CurrentTime) != X.GrabSuccess):
            print 'cannot grab the keyboard for the Menu!'
            self.close()
            return
        self.draw_menu()
        self.display.sync()
        self.used = False

    def expose_reply(self, _, event):
        if event.window == self.window:
            self.draw_menu()

    def motion_reply(self, _, event):
        if event.window != self.window:
            return

        # elghinn bug corrected =)
        if (not self.used and event.event_x == -1
            and event.event_y == -1):
            return

        self.used = True
        if (event.event_x < 0 or
            event.event_x > self.window.get_geometry().width):
            current = -1
        else:
            current = self.get_indice(event.event_y)
        if current != self.current:
            self.current = current
            self.draw_menu()

    def draw_menu(self, *args):
        window_geometry = self.window.get_geometry()
        width = window_geometry.width
        height = window_geometry.height
        #self.window_pixmap.fill_rectangle(self.black_gc, 0, 0, width, height)
        x, y = 0, 0
        for element in self.elements:
            if (self.get_indice(y) == self.current and
                element[1] != self.separator):
                fg_gc = self.black_gc
                bg_gc = self.darkgray_gc
            else:
                fg_gc = self.white_gc
                bg_gc = self.gray_gc

            if element[1] == self.separator and not element[0]:
                text_height = 2
                pixmap_font = self.window_pixmap.create_pixmap(
                    width, 4, self.screen.root_depth)
                pixmap_font.fill_rectangle(bg_gc, 0, 0, width, 4)
                pixmap_font.fill_rectangle(fg_gc, 1, 1, width - 2, 2)
                self.window_pixmap.copy_area(bg_gc, pixmap_font, 0, 0,
                                             width, 4, x, y)
                pixmap_font.free()
            else:
                text_height = self.font.query_text_extents(
                    element[0]).overall_ascent
                pixmap_font = self.window_pixmap.create_pixmap(
                    width, text_height + 2, self.screen.root_depth)
                pixmap_font.fill_rectangle(bg_gc, 0, 0, width, text_height + 2)
                pixmap_font.image_text(fg_gc, 1, text_height + 1, element[0])
                self.window_pixmap.copy_area(bg_gc, pixmap_font, 0, 0,
                                             width, text_height + 2, x, y)
                pixmap_font.free()

            y += text_height + 2
        self.window.copy_area(self.black_gc, self.window_pixmap, 0, 0,
                              window_geometry.width, window_geometry.height,
                              0, 0)
        self.display.sync()

    def button_release_reply(self, _, event):
        if not self.used or event.window != self.window:
            return

        # pointer outside of the menu x coordinates?
        if (event.event_x < 0 or
            event.event_x > self.window.get_geometry().width):
            self.close()
            return

        self.action(self.get_indice(event.event_y))

    def action(self, indice):
        if indice < 0 or indice >= len(self.elements):
            self.close()
            return

        self.close()

        element = self.elements[indice]
        if element[1] == self.separator:
            pass
        elif element[1] == self.python:
            element[2]()
        elif element[1] == self.shell:
            if len(element) == 4:
                self.wm.run_command(element[2], element[3])
            else:
                self.wm.run_command(element[2])
        elif element[1] == self.sub_menu:
            pass

    def key_press_reply(self, _, event):
        if event.window != self.window:
            return

        if event.detail == self.display.keysym_to_keycode(self.escape):
            self.close()
            return

        geo = self.window.get_geometry()
        #if (event.event_x < 0 or
        #    event.event_x > geo.width or
        #    event.event_y < 0 or
        #    event.event_y > geo.height):
        #    return

        pointer = self.window.query_pointer()
        if event.detail == self.display.keysym_to_keycode(self.up):
            if self.current == -1:
                self.wm.move_pointer(self.window, self.width/2,
                                     self.heights[0]/2)
            elif self.current == 0:
                self.wm.move_pointer(self.window, self.width/2,
                                     self.heigth - self.heights[-1]/2)
            elif self.current < len(self.elements):
                self.wm.move_pointer(self.window, self.width/2,
                                     sum(self.heights[:self.current - 1]) + \
                                     self.heights[self.current - 1]/2)
        elif event.detail in (self.display.keysym_to_keycode(self.down),
                              self.display.keysym_to_keycode(self.tab)):
            if self.current == -1:
                self.wm.move_pointer(self.window, self.width/2,
                                     self.heights[0] + self.heights[1]/2)
            elif self.current == len(self.elements) - 1:
                self.wm.move_pointer(self.window, self.width/2,
                                     self.heights[0]/2)
            elif self.current < len(self.elements) - 1:
                self.wm.move_pointer(self.window, self.width/2,
                                     sum(self.heights[:self.current + 1]) + \
                                     self.heights[self.current + 1]/2)
        elif event.detail == self.display.keysym_to_keycode(self.enter):
            self.action(self.current)

    def get_indice(self, y):
        if y < 0:
            return -1
        length = len(self.heights)
        for indice in xrange(length):
            if y < self.heights[indice]:
                return indice
            y -= self.heights[indice]
        return length

    def close(self):
        remove_pre_hook(self.wm.expose_reply, self.expose_reply)
        remove_pre_hook(self.wm.motion_reply, self.motion_reply)
        remove_pre_hook(self.wm.button_release_reply, self.button_release_reply)
        remove_pre_hook(self.wm.key_press_reply, self.key_press_reply)
        self.display.ungrab_pointer(X.CurrentTime)
        self.display.ungrab_keyboard(X.CurrentTime)
        self.window.set_wm_state(state=Xutil.WithdrawnState, icon=0)
        self.window.unmap()
        self.display.sync()
        self.window.destroy()
        self.window_pixmap.free()
        self.gray_gc.free()
        self.darkgray_gc.free()
        self.black_gc.free()
        self.white_gc.free()
        self.wm.set_new_focused_client(force=False)

class FocusesMenu(Menu):
    def __init__(self, wm, x, y):
        Menu.__init__(self, wm, x, y, self.build_menu_focuses(wm))

    @staticmethod
    def build_menu_focuses(wm):
        elements = [('Focuses Menu', Menu.separator),
                    ('', Menu.separator)]
        elements.extend([(focus, Menu.python, lambda f=focus: wm.set_focus(f))
                         for focus in wm.focuses])
        return elements

class IconifiedClientsMenu(Menu):
    def __init__(self, wm, x, y):
        Menu.__init__(self, wm, x, y, self.build_menu_iconified_clients(wm))

    @staticmethod
    def build_menu_iconified_clients(wm):
        elements = [('Iconified Clients Menu', Menu.separator),
                    ('', Menu.separator)]
        clients = [(client.name, Menu.python, client.uniconify)
                   for client in wm.clients
                   if client.states.icccm_state == client.iconified]
        if not clients:
            elements.append(('No iconified clients!', Menu.separator))
        else:
            elements.extend(clients)
        return elements

class ClientsSwitchMenu(Menu):
    def __init__(self, wm, shortcut, show_iconified_clients=False):
        mods = [modifiers[mod] for mod in shortcut.split('-')[:-1]
                if mod in modifiers]
        if good_key(shortcut)[0] == 0 or not mods:
            print 'ClientsSwitchMenu needs a good key shortcut and a modifier!'
            return
        self.show_iconified_clients = show_iconified_clients
        mappings = wm.display.get_modifier_mapping()
        self.modifiers = [mappings[modifier_indexes[mod]] for mod in mods]
        pointer = wm.root.query_pointer()
        self.old_x = pointer.root_x
        self.old_y = pointer.root_y
        self.putback_pointer = True
        Menu.__init__(self, wm, 0, 0, self.build_menu_clients(wm))
        if self.elements:
            # center the window and the pointer
            geo = self.window.get_geometry()
            self.window.configure(x=(wm.screen.width_in_pixels - geo.width)/2,
                                  y=(wm.screen.height_in_pixels - geo.height)/2)
            wm.move_pointer(self.window, geo.width/2, self.heights[0]/2)
            # check for key releases
            add_pre_hook(wm.key_release_reply, self.key_release_reply)

    def button_release_reply(self, _, event):
        self.putback_pointer = False
        Menu.button_release_reply(self, _, event)

    def key_release_reply(self, _, event):
        for modifier in self.modifiers[:]:
            for keycode in modifier:
                if event.detail == keycode:
                    self.modifiers.remove(modifier)
        if not self.modifiers:
            self.action(self.current)

    def switch_to_client(self, wm, client):
        def release_enter_reply(_, event):
            # we only care about the client where the pointer was in the end
            if event.window == blocked_client.parent:
                remove_pre_hook(wm.leave_reply, release_enter_reply)
                remove_pre_hook(wm.enter_reply, block_enter_reply)
                return
        def block_enter_reply(_, event):
            # do we need to block the event?
            if wm.get_managed_client(event.window) == blocked_client:
                # call the normal wm reply
                wm.enter_reply._wrapped_fun(wm, event)
                # discard the future calls
                event.window = 0
        if self.putback_pointer:
            wm.move_pointer(wm.root, self.old_x, self.old_y)
        if client.states.icccm_state == client.iconified:
            client.uniconify()
        client.upper()
        wm.give_focus(client)
        pointer = wm.root.query_pointer()
        blocked_client = wm.client_at_point(pointer.root_x, pointer.root_y)
        if blocked_client is not None and blocked_client != client:
            # block the EnterNotify events for the specified clients
            # until the pointer is moved away
            add_pre_hook(wm.enter_reply, block_enter_reply)
            add_pre_hook(wm.leave_reply, release_enter_reply)

    def close(self):
        if self.putback_pointer:
            self.wm.move_pointer(self.wm.root, self.old_x, self.old_y)
        Menu.close(self)
        remove_pre_hook(self.wm.key_release_reply, self.key_release_reply)

    def build_menu_clients(self, wm):
        clients = wm.clients[:]
        if clients and not clients[0].is_hidden():
            client = clients[0]
            clients.remove(client)
            clients.append(client)
            return [(client.name, Menu.python,
                     lambda wm=wm, client=client: self.switch_to_client(
                        wm, client)) for client in clients
                    if ((self.show_iconified_clients and
                         client.states.icccm_state == client.iconified) or
                        not client.is_hidden())]
